# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=missing-function-docstring,missing-param-doc,missing-raises-doc  # noqa: E501
"""Command line interface to convert or run a waldiez file."""

import json
import logging
import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Optional

import anyio
import typer
from typing_extensions import Annotated

from ._version import __version__
from .exporter import WaldiezExporter
from .models import Waldiez
from .runner import WaldiezRunner
from .utils import add_cli_extras

if TYPE_CHECKING:
    from autogen import ChatResult  # type: ignore[import-untyped]


app = typer.Typer(
    name="waldiez",
    help="Handle Waldiez flows.",
    context_settings={
        "help_option_names": ["-h", "--help"],
        "allow_extra_args": True,
        "ignore_unknown_options": True,
    },
    add_completion=False,
    no_args_is_help=True,
    invoke_without_command=True,
    add_help_option=True,
    pretty_exceptions_enable=False,
    epilog=("Use `waldiez [COMMAND] --help` for command-specific help. "),
)


@app.callback()
def show_version(
    version: bool = typer.Option(
        False,
        "--version",
        "-v",
        help="Show the version of the Waldiez package.",
    ),
) -> None:
    """Show the version of the Waldiez package and exit."""
    if version:
        typer.echo(f"waldiez version: {__version__}")
        raise typer.Exit()


@app.command()
def run(
    file: Annotated[
        Path,
        typer.Option(
            ...,
            help="Path to the Waldiez flow (*.waldiez) file.",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            resolve_path=True,
        ),
    ],
    output: Optional[Path] = typer.Option(
        None,
        help=(
            "Path to the output (.py) file. "
            "The output's directory will contain "
            "the generated flow (.py) and any additional generated files."
        ),
        dir_okay=False,
        resolve_path=True,
    ),
    force: bool = typer.Option(
        False,
        help="Override the output file if it already exists.",
    ),
) -> None:
    """Run a Waldiez flow."""
    # a swarm chat without a user agent
    # creates a new user (this has a default code execution with docker)
    # temp (until we handle/detect docker setup)
    os.environ["AUTOGEN_USE_DOCKER"] = "0"
    output_path = _get_output_path(output, force)
    with file.open("r", encoding="utf-8") as _file:
        try:
            data = json.load(_file)
        except json.decoder.JSONDecodeError as error:
            typer.echo("Invalid .waldiez file. Not a valid json?")
            raise typer.Exit(code=1) from error
    waldiez = Waldiez.from_dict(data)
    runner = WaldiezRunner(waldiez)
    if waldiez.is_async:
        results = anyio.run(runner.a_run, output_path)
    else:
        results = runner.run(output_path=output_path)
    logger = _get_logger()
    if isinstance(results, list):
        logger.info("Results:")
        for result in results:
            _log_result(result, logger)
            sep = "-" * 80
            print("\n" + f"{sep}" + "\n")
    elif isinstance(results, dict):
        logger.info("Results:")
        for key, result in results.items():
            logger.info("Order: %s", key)
            _log_result(result, logger)
            sep = "-" * 80
            print("\n" + f"{sep}" + "\n")
    else:
        _log_result(results, logger)


@app.command()
def convert(
    file: Annotated[
        Path,
        typer.Option(
            ...,
            help="Path to the Waldiez flow (*.waldiez) file.",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            resolve_path=True,
        ),
    ],
    output: Annotated[
        Path,
        typer.Option(
            ...,
            help=(
                "Path to the output file. "
                "The file extension determines the output format: "
                "`.py` for Python script, `.ipynb` for Jupyter notebook."
            ),
            file_okay=True,
            dir_okay=False,
            resolve_path=True,
        ),
    ],
    force: bool = typer.Option(
        False,
        help="Override the output file if it already exists.",
    ),
) -> None:
    """Convert a Waldiez flow to a Python script or a Jupyter notebook."""
    _get_output_path(output, force)
    with file.open("r", encoding="utf-8") as _file:
        try:
            data = json.load(_file)
        except json.decoder.JSONDecodeError as error:
            typer.echo("Invalid .waldiez file. Not a valid json?")
            raise typer.Exit(code=1) from error
    waldiez = Waldiez.from_dict(data)
    exporter = WaldiezExporter(waldiez)
    exporter.export(output, force=force)
    generated = str(output).replace(os.getcwd(), ".")
    typer.echo(f"Generated: {generated}")


@app.command()
def check(
    file: Annotated[
        Path,
        typer.Option(
            ...,
            help="Path to the Waldiez flow (*.waldiez) file.",
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            resolve_path=True,
        ),
    ],
) -> None:
    """Validate a Waldiez flow."""
    with file.open("r", encoding="utf-8") as _file:
        data = json.load(_file)
    Waldiez.from_dict(data)
    typer.echo("Waldiez flow is valid.")


def _get_output_path(output: Optional[Path], force: bool) -> Optional[Path]:
    if output is not None:
        output = Path(output).resolve()
    if output is not None and not output.parent.exists():
        output.parent.mkdir(parents=True)
    if output is not None and output.exists():
        if force is False:
            typer.echo("Output file already exists.")
            raise typer.Exit(code=1)
        output.unlink()
    return output


def _get_logger(level: int = logging.INFO) -> logging.Logger:
    """Get the logger for the Waldiez package.

    Parameters
    ----------
    level : int or str, optional
        The logging level. Default is logging.INFO.

    Returns
    -------
    logging.Logger
        The logger.
    """
    # check if we already have setup a config

    if not logging.getLogger().handlers:
        logging.basicConfig(
            level=level,
            format="%(levelname)s %(message)s",
            stream=sys.stderr,
            force=True,
        )
    logger = logging.getLogger("waldiez::cli")
    current_level = logger.getEffectiveLevel()
    if current_level != level:
        logger.setLevel(level)
    return logger


def _log_result(result: "ChatResult", logger: logging.Logger) -> None:
    """Log the result of the Waldiez flow."""
    logger.info("Chat History:\n")
    logger.info(result.chat_history)
    logger.info("Summary:\n")
    logger.info(result.summary)
    logger.info("Cost:\n")
    logger.info(result.cost)


add_cli_extras(app)

if __name__ == "__main__":
    _get_logger()
    app()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Waldiez package."""

from ._version import __version__
from .exporter import WaldiezExporter
from .models import Waldiez
from .runner import WaldiezRunner
from .utils import check_conflicts, check_flaml_warnings

__WALDIEZ_INITIALIZED = False

if not __WALDIEZ_INITIALIZED:
    check_conflicts()
    check_flaml_warnings()
    # let's skip the one below
    # check_pysqlite3()
    # and use it only if needed:
    #   captain_agent dependency:
    #   before calling pip install pyautogen[captainagent]
    #   we should have pysqlite3 installed (at least on windows)
    # before running a flow
    __WALDIEZ_INITIALIZED = True

__all__ = [
    "Waldiez",
    "WaldiezExporter",
    "WaldiezRunner",
    "__version__",
]



"""Dictionary related utilities."""

from typing import Any, Dict


def update_dict(original: Dict[str, Any]) -> Dict[str, Any]:
    """

    Try to determine the type of the dictionary values.

    Parameters
    ----------
    original : Dict[str, Any]
        The original dictionary.

    Returns
    -------
    Dict[str, Any]
        The updated dictionary with values converted to the detected types.
    """
    new_dict: Dict[str, Any] = {}
    for key, value in original.items():
        value_lower = str(value).lower()
        if value_lower in ("none", "null"):
            new_dict[key] = None
        elif value_lower in ("true", "false"):
            new_dict[key] = value_lower == "true"
        elif str(value).isdigit():
            new_dict[key] = int(value)
        elif str(value).replace(".", "").isdigit():
            try:
                new_dict[key] = float(value)
            except ValueError:  # pragma: no cover
                new_dict[key] = value
        else:
            new_dict[key] = value
    return new_dict



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""
The role of the exporter is to export the model's data
to an autogen's flow with one or more chats.

The resulting file(s): a `flow.py` file with one `main()` function
to trigger the chat(s).
If additional tools/skills are used,
they are exported as their `skill_name` in the same directory with
the `flow.py` file. So the `flow.py` could have entries like:
`form {flow_name}_{skill1_name} import {skill1_name}`
`form {flow_name}_{skill2_name} import {skill2_name}`
"""

# pylint: disable=inconsistent-quotes

import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import List, Optional, Union

from .exporting import FlowExporter
from .models import Waldiez


class WaldiezExporter:
    """Waldiez exporter.

    Attributes:
        waldiez (Waldiez): The Waldiez instance.
    """

    def __init__(self, waldiez: Waldiez) -> None:
        """Initialize the Waldiez exporter.

        Parameters:
            waldiez (Waldiez): The Waldiez instance.
        """
        self.waldiez = waldiez
        # self._initialize()

    @classmethod
    def load(cls, file_path: Path) -> "WaldiezExporter":
        """Load the Waldiez instance from a file.

        Parameters
        ----------
        file_path : Path
            The file path.

        Returns
        -------
        WaldiezExporter
            The Waldiez exporter.
        """
        waldiez = Waldiez.load(file_path)
        return cls(waldiez)

    def export(self, path: Union[str, Path], force: bool = False) -> None:
        """Export the Waldiez instance.

        Parameters
        ----------
        path : Union[str, Path]
            The path to export to.
        force : bool, optional
            Override the output file if it already exists, by default False.

        Raises
        ------
        FileExistsError
            If the file already exists and force is False.
        IsADirectoryError
            If the output is a directory.
        ValueError
            If the file extension is invalid.
        """
        if not isinstance(path, Path):
            path = Path(path)
        path = path.resolve()
        if path.is_dir():
            raise IsADirectoryError(f"Output is a directory: {path}")
        if path.exists():
            if force is False:
                raise FileExistsError(f"File already exists: {path}")
            path.unlink(missing_ok=True)
        path.parent.mkdir(parents=True, exist_ok=True)
        extension = path.suffix
        if extension == ".waldiez":
            self.to_waldiez(path)
        elif extension == ".py":
            self.to_py(path)
        elif extension == ".ipynb":
            self.to_ipynb(path)
        else:
            raise ValueError(f"Invalid extension: {extension}")

    def to_ipynb(self, path: Path) -> None:
        """Export flow to jupyter notebook.

        Parameters
        ----------
        path : Path
            The path to export to.

        Raises
        ------
        RuntimeError
            If the notebook could not be generated.
        """
        # we first create a .py file with the content
        # and then convert it to a notebook using jupytext
        exporter = FlowExporter(
            waldiez=self.waldiez,
            output_dir=path.parent,
            for_notebook=True,
        )
        output = exporter.export()
        content = output["content"]
        if not content:
            raise RuntimeError("Could not generate notebook")
        py_path = path.with_suffix(".tmp.py")
        with open(py_path, "w", encoding="utf-8", newline="\n") as f:
            f.write(content)
        if not shutil.which("jupytext"):  # pragma: no cover
            run_command(
                [sys.executable, "-m", "pip", "install", "jupytext"],
                allow_error=False,
            )
        run_command(
            [
                sys.executable,
                "-m",
                "jupytext",
                "--to",
                "notebook",
                str(py_path),
            ],
            allow_error=False,
        )
        ipynb_path = str(py_path).replace(".tmp.py", ".tmp.ipynb")
        if not os.path.exists(ipynb_path):  # pragma: no cover
            raise RuntimeError("Could not generate notebook")
        Path(ipynb_path).rename(ipynb_path.replace(".tmp.ipynb", ".ipynb"))
        py_path.unlink(missing_ok=True)

    def to_py(self, path: Path) -> None:
        """Export waldiez flow to python script.

        Parameters
        ----------
        path : Path
            The path to export to.

        Raises
        ------
        RuntimeError
            If the python script could not be generated.
        """
        exporter = FlowExporter(
            waldiez=self.waldiez,
            output_dir=path.parent,
            for_notebook=False,
        )
        output = exporter.export()
        content = output["content"]
        if not content:
            raise RuntimeError("Could not generate python script")
        with open(path, "w", encoding="utf-8", newline="\n") as file:
            file.write(content)

    def to_waldiez(self, file_path: Path) -> None:
        """Export the Waldiez instance.

        Parameters
        ----------
        file_path : Path
            The file path.
        """
        with open(file_path, "w", encoding="utf-8", newline="\n") as file:
            file.write(self.waldiez.model_dump_json())


def run_command(
    cmd: List[str],
    cwd: Optional[Path] = None,
    allow_error: bool = True,
) -> None:
    """Run a command.

    Parameters
    ----------
    cmd : List[str]
        The command to run.
    cwd : Path, optional
        The working directory, by default None (current working directory).
    allow_error : bool, optional
        Whether to allow errors, by default True.

    Raises
    ------
    RuntimeError
        If the command fails and allow_error is False.
    """
    if not cwd:
        cwd = Path.cwd()
    # pylint: disable=broad-except
    try:
        subprocess.run(
            cmd,
            check=True,
            cwd=cwd,
            env=os.environ,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )  # nosemgrep # nosec
    except BaseException as error:  # pragma: no cover
        if allow_error:
            return
        raise RuntimeError(f"Error running command: {error}") from error



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Helpers for the flow model."""

import uuid
from datetime import datetime, timezone
from typing import List

from ..agents import (
    WaldiezAgent,
    WaldiezAgentNestedChat,
    WaldiezAgentNestedChatMessage,
    WaldiezSwarmAgent,
    WaldiezSwarmOnCondition,
)
from ..chat import WaldiezChat


def id_factory() -> str:
    """Generate a unique ID.

    Returns
    -------
    str
        The unique ID.
    """
    now_td = datetime.now(timezone.utc)
    now_str = now_td.strftime("%Y%m%d%H%M%S%f")
    return f"{now_str}-{uuid.uuid4().hex}"


def check_handoff_to_nested_chat(
    agent: WaldiezSwarmAgent,
    all_agents: List[WaldiezAgent],
    all_chats: List[WaldiezChat],
) -> None:
    """Check the handoffs to a nested chat.

    If we have one and the agent does not have nested_chats,
    we should generate them with the `handoff.target.id`
    as the first (chat's) message.

    Parameters
    ----------
    agent : WaldiezSwarmAgent
        The swarm agent.
    all_agents : List[WaldiezAgent]
        All agents.
    all_chats : List[WaldiezChat]
        All chats.

    Raises
    ------
    ValueError
        If the agent has a handoff to a nested chat,
        but no chat found with it as a source.
    """
    # pylint: disable=too-complex
    for handoff in agent.handoffs:
        if not isinstance(handoff, WaldiezSwarmOnCondition):
            continue
        is_nested_chat = handoff.target_type == "nested_chat"
        if is_nested_chat:
            # check if the agent already has nested_chats
            # but only to get the order (and the first chat/message)
            # either way, we must include all the connections that
            # are swarm => non-swarm
            # if we have the orders, ok, else we get them from the
            # edges positions
            all_connections = sorted(
                get_nested_chat_swarm_connections(agent, all_agents, all_chats),
                key=lambda x: x.data.order,
            )
            if not all_connections:
                raise ValueError(
                    f"Agent {agent.name} has a handoff to a nested chat, "
                    "but no chat found with it as a source."
                )
            agent_nested_chats = agent.nested_chats
            if not agent_nested_chats or not agent_nested_chats[0].messages:
                agent.data.nested_chats = [
                    WaldiezAgentNestedChat(
                        triggered_by=[],
                        messages=[
                            WaldiezAgentNestedChatMessage(
                                id=chat.id,
                                is_reply=False,
                            )
                            for chat in all_connections
                        ],
                    )
                ]
                break
            agent.data.nested_chats = merge_nested_chat_messages(
                agent_nested_chats[0].messages, all_connections
            )


def get_nested_chat_swarm_connections(
    agent: WaldiezSwarmAgent,
    all_agents: List[WaldiezAgent],
    all_chats: List[WaldiezChat],
) -> List[WaldiezChat]:
    """Get the nested chat connections.

    Parameters
    ----------
    agent : WaldiezSwarmAgent
        The swarm agent.
    all_agents : Iterator[WaldiezAgent]
        All agents.
    all_chats : List[WaldiezChat]
        All chats.

    Returns
    -------
    List[WaldiezAgentNestedChat]
        The nested chat connections.
    """
    connections_with_non_swarm_targets = []
    for chat in all_chats:
        if chat.source != agent.id:
            continue
        target_agent = next(
            (a for a in all_agents if a.id == chat.target), None
        )
        if not target_agent or target_agent.agent_type != "swarm":
            connections_with_non_swarm_targets.append(chat)
    return connections_with_non_swarm_targets


def merge_nested_chat_messages(
    agent_nested_chat_messages: List[WaldiezAgentNestedChatMessage],
    all_connections: List[WaldiezChat],
) -> List[WaldiezAgentNestedChat]:
    """Merge the nested chat messages.

    Parameters
    ----------
    all_connections : List[WaldiezChat]
        The connections.
    agent_nested_chat_messages : List[WaldiezAgentNestedChatMessage]
        The agent's nested chat messages.

    Returns
    -------
    List[WaldiezAgentNestedChat]
        The merged nested chat with all the messages.
    """
    nested_chat = WaldiezAgentNestedChat(triggered_by=[], messages=[])
    chat_ids_added: List[str] = []
    for message in agent_nested_chat_messages:
        chat = next((c for c in all_connections if c.id == message.id), None)
        if chat and chat.id not in chat_ids_added:
            nested_chat.messages.append(
                WaldiezAgentNestedChatMessage(
                    id=chat.id,
                    is_reply=False,
                )
            )
            chat_ids_added.append(chat.id)
    for chat in all_connections:
        if chat.id not in chat_ids_added:
            nested_chat.messages.append(
                WaldiezAgentNestedChatMessage(
                    id=chat.id,
                    is_reply=False,
                )
            )
            chat_ids_added.append(chat.id)
    nested_chat.messages.sort(key=lambda x: chat_ids_added.index(x.id))
    return [nested_chat]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Flow Exporter class.

- We gather all the exports (models, skills, agents, chats).

- We first add all the imports from the above exports.
- If we have skills, we include their imports.
    (their files were generated when exporting the skills).
- Then, we write the all model configs.
- Next, we write the agent definitions
    (using the `llm_config=...` argument from the model exports).
- If additional (nested_chats) are defined,
    we write their registrations after all agents are defined.
- Next, we write the chat definitions
    (using the agent names from the agent exports).
- If exporting to py,
    we add the `run` function and the `def main()` to call the run function.
- If the flow is async, the `run` function is async.
"""

# flake8: noqa E501
# pylint: disable=line-too-long

from functools import partial
from pathlib import Path
from typing import List, Optional, Tuple, Union

from waldiez.models import Waldiez, WaldiezAgent

from ..agent import AgentExporter
from ..base import (
    AgentPosition,
    AgentPositions,
    BaseExporter,
    ExporterMixin,
    ExporterReturnType,
    ExportPosition,
    ExportPositions,
    ImportPosition,
)
from ..chats import ChatsExporter
from ..models import ModelsExporter
from ..skills import SkillsExporter
from .utils import (
    add_after_agent_content,
    add_after_chat_content,
    add_before_agent_content,
    ensure_unique_names,
    gather_agent_outputs,
    gather_imports,
    get_after_run_content,
    get_def_main,
    get_ipynb_content_start,
    get_py_content_start,
    get_sqlite_out,
    get_start_logging,
    get_stop_logging,
    get_the_imports_string,
)


# pylint: disable=too-many-instance-attributes
class FlowExporter(BaseExporter, ExporterMixin):
    """Flow exporter."""

    def __init__(
        self,
        waldiez: Waldiez,
        for_notebook: bool,
        output_dir: Optional[Union[str, Path]] = None,
    ) -> None:
        """Initialize the flow exporter."""
        self.waldiez = waldiez
        self.for_notebook = for_notebook
        if output_dir is not None and not isinstance(output_dir, Path):
            output_dir = Path(output_dir).resolve()
        self.output_dir = output_dir
        self.initialize()

    def initialize(
        self,
    ) -> None:
        """Get all the names in the flow.

        We need to make sure that no duplicate names are used,
        and that the names can be used as python variables.
        """
        unique_names = ensure_unique_names(
            self.waldiez,
            self.get_valid_instance_name,
        )
        self.flow_name = unique_names["flow_name"]
        self.agents = unique_names["agents"]
        self.models = unique_names["models"]
        self.skills = unique_names["skills"]
        self.chats = unique_names["chats"]
        self.agent_names = unique_names["agent_names"]
        self.model_names = unique_names["model_names"]
        self.skill_names = unique_names["skill_names"]
        self.chat_names = unique_names["chat_names"]

    def export_flow(self) -> ExporterReturnType:
        """Export the flow.

        Returns
        -------
        ExporterReturnType
            The exported flow.
        """
        models_output = self.export_models()
        skills_output = self.export_skills()
        chats_output = self.export_chats()
        env_vars = self.gather_environment_variables(
            model_env_vars=models_output["environment_variables"],
            skill_env_vars=skills_output["environment_variables"],
            chat_env_vars=chats_output["environment_variables"],
        )
        before_export = self.gather_exports(
            model_export=models_output["before_export"],
            skill_export=skills_output["before_export"],
            chat_export=chats_output["before_export"],
        )
        after_export = self.gather_exports(
            model_export=models_output["after_export"],
            skill_export=skills_output["after_export"],
            chat_export=chats_output["after_export"],
        )
        # agents last (to make sure we have any needed arguments)
        # like `llm_config=...` from the models
        agents_output = self.export_agents(
            before_export=before_export,
            after_export=after_export,
        )
        imports = gather_imports(
            model_imports=models_output["imports"],
            skill_imports=skills_output["imports"],
            chat_imports=chats_output["imports"],
            agent_imports=agents_output["imports"],
        )
        if agents_output["environment_variables"]:
            env_vars.extend(agents_output["environment_variables"])
        if agents_output["before_export"]:
            before_export.extend(agents_output["before_export"])
        if agents_output["after_export"]:
            after_export.extend(agents_output["after_export"])
        all_imports = (
            get_the_imports_string(imports, is_async=self.waldiez.is_async),
            ImportPosition.LOCAL,
        )
        before_chats_export = chats_output["before_export"] or []
        content_before_chats = [
            x[0]
            for x in before_chats_export
            if isinstance(x[1], ExportPosition)
            and x[1].position == ExportPositions.CHATS
        ]
        before_chats = "\n".join(content_before_chats)
        content = self.merge_exports(
            imports=all_imports,
            models_output=models_output["content"] or "",
            skills_output=skills_output["content"] or "",
            agents_content=agents_output["content"] or "",
            chats_content=chats_output["content"] or "",
            before_chats=before_chats,
        )
        return {
            "content": content,
            "imports": [all_imports],
            "after_export": after_export,
            "before_export": before_export,
            "environment_variables": env_vars,
        }

    def merge_exports(
        self,
        imports: Tuple[str, ImportPosition],
        models_output: str,
        skills_output: str,
        agents_content: str,
        chats_content: str,
        before_chats: str,
    ) -> str:
        """Merge all the export contents.

        Parameters
        ----------
        imports : Tuple[str, ImportPosition]
            The imports.
        models_output : str
            The models output.
        skills_output : str
            The skills output.
        agents_content : str
            The agents content.
        chats_content : str
            The chats content.
        before_chats : str
        Returns
        -------
        str
            The merged export contents.
        """
        is_async = self.waldiez.is_async
        content = (
            get_py_content_start(self.waldiez)
            if not self.for_notebook
            else get_ipynb_content_start(self.waldiez, comment=self.comment)
        )
        content += self.get_comment("imports", self.for_notebook) + "\n"
        content += imports[0] + "\n"
        content += self.get_comment("logging", self.for_notebook) + "\n"
        content += get_start_logging(tabs=0) + "\n"
        content += "start_logging()\n\n"
        if models_output:
            content += self.get_comment("models", self.for_notebook) + "\n"
            content += models_output + "\n"
        if skills_output:
            content += self.get_comment("skills", self.for_notebook) + "\n"
            content += skills_output + "\n"
        if agents_content:
            content += self.get_comment("agents", self.for_notebook) + "\n"
            content += agents_content + "\n"
        if before_chats:
            content += before_chats + "\n"
        content += get_sqlite_out(is_async=is_async) + "\n"
        content += get_stop_logging(tabs=0, is_async=is_async) + "\n"
        content += self.get_comment("run", self.for_notebook) + "\n"
        after_run = get_after_run_content(
            waldiez=self.waldiez,
            agent_names=self.agent_names,
            tabs=0 if self.for_notebook else 1,
        )
        if self.for_notebook is False:
            content += get_def_main(
                chats_content,
                after_run=after_run,
                is_async=self.waldiez.is_async,
            )
        else:
            content += "\n" + chats_content + "\n"
            if is_async:
                content += "await stop_logging()"
            else:
                content += "stop_logging()"
                content += after_run
        content = content.replace("\n\n\n\n", "\n\n\n")
        return content

    @staticmethod
    def gather_environment_variables(
        model_env_vars: Optional[List[Tuple[str, str]]],
        skill_env_vars: Optional[List[Tuple[str, str]]],
        chat_env_vars: Optional[List[Tuple[str, str]]],
    ) -> List[Tuple[str, str]]:
        """
        Gather all the environment variables.

        Parameters
        ----------
        model_env_vars : Optional[List[Tuple[str, str]]]
            The model environment variables.
        skill_env_vars : Optional[List[Tuple[str, str]]]
            The skill environment variables.
        chat_env_vars : Optional[List[Tuple[str, str]]]
            The chat environment variables.

        Returns
        -------
        List[Tuple[str, str]]
            The gathered environment variables.
        """
        all_env_vars: List[Tuple[str, str]] = []
        if model_env_vars:
            all_env_vars.extend(model_env_vars)
        if skill_env_vars:
            all_env_vars.extend(skill_env_vars)
        if chat_env_vars:
            all_env_vars.extend(chat_env_vars)
        return all_env_vars

    @staticmethod
    def gather_exports(
        model_export: Optional[
            List[Tuple[str, Union[ExportPosition, AgentPosition]]]
        ],
        skill_export: Optional[
            List[Tuple[str, Union[ExportPosition, AgentPosition]]]
        ],
        chat_export: Optional[
            List[Tuple[str, Union[ExportPosition, AgentPosition]]]
        ],
    ) -> List[Tuple[str, Union[ExportPosition, AgentPosition]]]:
        """Gather all (but agents) the before or after exports.

        Parameters
        ----------
        model_export : Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]
            The model exports.
        skill_export : Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]
            The skill exports.
        chat_export : Optional[List[Tuple[str, Union[ExportPosition, AgentPosition]]]]
            The chat exports.

        Returns
        -------
        List[Tuple[str, Union[ExportPosition, AgentPosition]]]
            The gathered exports.
        """
        all_exports: List[Tuple[str, Union[ExportPosition, AgentPosition]]] = []
        if model_export:
            all_exports.extend(model_export)
        if skill_export:
            all_exports.extend(skill_export)
        if chat_export:
            all_exports.extend(chat_export)
        return all_exports

    def export_models(self) -> ExporterReturnType:
        """Export the models.

        Returns
        -------
        str
            The exported models.
        """
        exporter = ModelsExporter(
            flow_name=self.flow_name,
            agents=self.agents,
            agent_names=self.agent_names,
            models=self.models,
            model_names=self.model_names,
            for_notebook=self.for_notebook,
            output_dir=self.output_dir,
        )
        return exporter.export()

    def export_skills(self) -> ExporterReturnType:
        """Export the skills.

        Returns
        -------
        str
            The exported skills.
        """
        exporter = SkillsExporter(
            flow_name=self.flow_name,
            agents=self.agents,
            agent_names=self.agent_names,
            skills=self.skills,
            skill_names=self.skill_names,
            output_dir=self.output_dir,
        )
        return exporter.export()

    @staticmethod
    def gather_agent_arguments(
        before_export: List[Tuple[str, Union[ExportPosition, AgentPosition]]],
        after_export: List[Tuple[str, Union[ExportPosition, AgentPosition]]],
    ) -> List[Tuple[str, AgentPosition]]:
        """Gather the agent arguments.

        Parameters
        ----------
        before_export : List[Tuple[str, Union[ExportPosition, AgentPosition]]]
            The before export.
        after_export : List[Tuple[str, Union[ExportPosition, AgentPosition]]]
            The after export.

        Returns
        -------
        List[Tuple[str, AgentPosition]]
            The gathered agent arguments.
        """
        exported_with_agent_arg: List[Tuple[str, AgentPosition]] = []
        for before in before_export:
            position = before[1]
            if (
                isinstance(position, AgentPosition)
                and position.position == AgentPositions.AS_ARGUMENT
            ):
                exported_with_agent_arg.append((before[0], position))
        for after in after_export:
            position = after[1]
            if (
                isinstance(position, AgentPosition)
                and position.position == AgentPositions.AS_ARGUMENT
            ):
                exported_with_agent_arg.append((after[0], position))
        return exported_with_agent_arg

    def export_agents(
        self,
        before_export: List[Tuple[str, Union[ExportPosition, AgentPosition]]],
        after_export: List[Tuple[str, Union[ExportPosition, AgentPosition]]],
    ) -> ExporterReturnType:
        """Export the agents.

        Parameters
        ----------
        before_export : List[Tuple[str, Union[ExportPosition, AgentPosition]]]
            The before export.
        after_export : List[Tuple[str, Union[ExportPosition, AgentPosition]]]
            The after export.
        Returns
        -------
        str
            The exported agents.
        """
        agent_outputs = []
        for agent in self.agents:
            exported_with_agent_arg = self.gather_agent_arguments(
                before_export, after_export
            )
            arguments_resolver = partial(
                self.agent_arguments_resolver,
                exported_with_agent_arg,
            )
            group_chat_members = self.waldiez.get_group_chat_members(agent)
            exporter = AgentExporter(
                agent=agent,
                agent_names=self.agent_names,
                model_names=self.model_names,
                skill_names=self.skill_names,
                chats=(self.chats, self.chat_names),
                is_async=self.waldiez.is_async,
                for_notebook=self.for_notebook,
                output_dir=self.output_dir,
                group_chat_members=group_chat_members,
                arguments_resolver=arguments_resolver,
            )
            agent_output = exporter.export()
            agent_content = agent_output["content"] or ""
            after_agent_export = agent_output["after_export"]
            if after_agent_export:
                after_export.extend(after_agent_export)
            before_agent_export = agent_output["before_export"]
            if before_agent_export:
                before_export.extend(before_agent_export)
            if agent_content:
                agent_content = add_before_agent_content(
                    agent_content,
                    before_export,
                    agent,
                )
                agent_content = add_after_agent_content(
                    agent_content,
                    after_export,
                    agent,
                )
            agent_output["content"] = agent_content
            agent_outputs.append(agent_output)
        return gather_agent_outputs(
            before_export=before_export,
            after_export=after_export,
            agent_outputs=agent_outputs,
        )

    @staticmethod
    def agent_arguments_resolver(
        additional_exports: List[Tuple[str, AgentPosition]], agent: WaldiezAgent
    ) -> List[str]:
        """Resolve the arguments for the agent.

        Parameters
        ----------
        additional_exports : List[Tuple[str, AgentPosition]]
            The additional exports.
        agent : WaldiezAgent
            The agent.

        Returns
        -------
        List[str]
            The arguments for the agent.
        """
        return [x[0] for x in additional_exports if x[1].agent == agent]

    def export_chats(self) -> ExporterReturnType:
        """Export the chats.

        Returns
        -------
        str
            The exported chats.
        """
        exporter = ChatsExporter(
            get_swarm_members=self.waldiez.get_swarm_members,
            all_agents=self.agents,
            agent_names=self.agent_names,
            all_chats=self.chats,
            chat_names=self.chat_names,
            main_chats=self.waldiez.chats,
            for_notebook=self.for_notebook,
            is_async=self.waldiez.is_async,
        )
        output = exporter.export()
        chat_contents = output["content"] or ""
        after_chats = output["after_export"]
        if chat_contents and after_chats:
            chat_contents = add_after_chat_content(
                chat_contents,
                after_chats,
            )
        output["content"] = chat_contents
        return output

    def export(self) -> ExporterReturnType:
        """Export the flow.

        Returns
        -------
        SubExporterReturnType
            The exported flow.
        """
        return self.export_flow()
